######################################################################
# Copyright (c) 2018 Josef Spidlen and Christoph Freier, FlowJo, LLC
#
# License
# The software is distributed under the terms of the
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
#
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct,
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this
# software.
######################################################################

suppressMessages(library("flowCore"))
suppressMessages(library("tidyverse"))

# For debug purposes
sessionInfo()
## Just for debugging to see what FlowJo is actually giving us
# cat("FJ-DATA-FILE-PATH    = FJ_DATA_FILE_PATH \n")
# cat("FJ-CSV-IDX-FILE-PATH = FJ_CSV_IDX_FILE_PATH \n")
# cat("FJ-CSV-OUPUT-FILE    = FJ_CSV_OUPUT_FILE \n")

##### Inputs
# Input fcs
inputFCS <- read.FCS("FJ_DATA_FILE_PATH", emptyValue = FALSE, truncate_max_range = FALSE)
## Error message if unable to read the fcs file correctly
if(!nrow(inputFCS) > 0)
	stop("Unable to read the input FCS file correctly.", call. = FALSE)

##### Sanity Checks
# Currently supported machines
supCYT <- c("FACSAriaIII",
		"FACSAria Fusion (FACSAriaIII)",
		"BD FACSymphony S6",
		"BD FACSMelody",
		"BD Influx System (USB)",
		"MoFlo Astrios")

# Check $CYT
cytKey <- "FJ_CYT_PAR"
## If auto-selection then take it from $CYT
if (cytKey == "Auto") {
	cytKey <- inputFCS@description$`$CYT`
	## Error message if input FCS file doesn't specify the $CYT
	if(length(cytKey) == 0)
		stop("The input FCS file doesn't specify the $CYT.", call. = FALSE)
	## Error message if cell sorter is not currently supported
	if(!(cytKey %in% supCYT))
		stop("This cell sorter is not currently supported.", call. = FALSE)
}

##### BD FACSAriaIII, FACSAria Fusion & S6
if (cytKey %in% supCYT[1:3]) {
	# Find the index sorting locations
	indSor <- rev(grep("INDEX SORTING LOCATIONS", names(inputFCS@description))) %>%
			sapply(., function(x) unlist(inputFCS@description[[x]])) %>%
			paste(collapse = "") %>%
			str_split(";") %>%
			unlist %>%
			.[. != ""]
	
	## Error message if unable to find the sorting locations
	if(length(indSor) == 0)
		stop("Unable to find the index sorting locations.", call. = FALSE)
	
	# Generate results
	results <- data.frame(IdxRow = as.integer(gsub("([0-9]?[0-9]),([0-9]?[0-9])", "\\1", indSor)),
					IdxCol = as.integer(gsub("([0-9]?[0-9]),([0-9]?[0-9])", "\\2", indSor))) %>%
			mutate(IdxRow = rawToChar(as.raw(65 + as.integer(.[, "IdxRow"])), multiple = TRUE)) %>%
			mutate(IdxCol = formatC((IdxCol + 1), width = 2, flag = 0))
	
	# Generate perCell
	perCell <- cbind(data.frame(inputFCS@exprs), results) %>%
			mutate(Well = paste0(IdxRow, IdxCol))
	
	##### BD FACSMelody
} else if (cytKey %in% supCYT[4]) {
	# Input csv
	inputCSV <- read.csv("FJ_CSV_IDX_FILE_PATH")
	## Error message if unable to read the fcs file correctly
	if(!nrow(inputCSV) > 0)
		stop("Unable to read the input CSV file correctly.", call. = FALSE)
	
	# Correct csv
	## Remove header
	inputCSV <- read.csv("FJ_CSV_IDX_FILE_PATH", skip = which(inputCSV[, 1] == "Well")) %>% 
			mutate_at(vars(FSC.A:SSC.W), round)
	
	# Generate perCell
	perCell <- round(data.frame(inputFCS@exprs)) %>%
			select(Time, matches("[FS]SC.[AHW]")) %>%
			merge(., inputCSV, all.x = TRUE) %>%
			mutate(Well, IdxRow = gsub("([[:alpha:]])([0-9][0-9])", "\\1", Well)) %>% ##### Review
			mutate(Well, IdxCol = gsub("([[:alpha:]])([0-9][0-9])", "\\2", Well))     ##### this!!
	
	# Generate results
	results <- perCell %>%
			select(IdxRow, IdxCol)
	
	# Correct perCell
	perCell <- perCell %>%
			filter(Well != "NA") %>%
			arrange(Well)
	
	##### BD Influx System
} else if (cytKey %in% supCYT[5]) {
	# Find the index sorting positions
	indSor <- rev(grep("INDEXSORTPOSITIONS", names(inputFCS@description))) %>%
			sapply(., function(x) unlist(inputFCS@description[[x]])) %>%
			paste(collapse = "") %>%
			str_split(",") %>%
			unlist %>%
			matrix(ncol = 3, byrow = TRUE)
	## Error message if unable to find the sorting positions
	if(length(indSor) == 0)
		stop("Unable to find the index sorting positions.", call. = FALSE)
	
	# Round Tray.X and Tray.Y
	trayXY <- data.frame(Well = indSor[, 1],
			Tray.X = indSor[, 2] <- round(as.numeric(indSor[, 2])/100),
			Tray.Y = indSor[, 3] <- round(as.numeric(indSor[, 3])/100),
			Sort.Result.Bits = 32769)
	
	# Generate perCell
	perCell <- data.frame(inputFCS@exprs) %>%
			mutate(Tray.X = round(Tray.X/100),
					Tray.Y = round(Tray.Y/100)) %>% 
			merge(., trayXY, all.x = TRUE) %>% 
			mutate(IdxRow = gsub("([[:alpha:]])([0-9]?[0-9])", "\\1", Well),
					IdxCol = gsub("([[:alpha:]])([0-9]?[0-9])", "\\2", Well)) %>% 
			mutate(IdxCol = formatC((as.numeric(IdxCol)), width = 2, flag = 0))
	
	# Generate results
	results <- perCell %>%
			select(IdxRow, IdxCol)
	
	# Correct perCell
	perCell <- perCell %>%
			filter(IdxRow != "NA") %>%
			mutate(Well = paste0(IdxRow, IdxCol)) %>%
			arrange(Well)
	
	##### MoFlo Astrios
} else if (cytKey %in% supCYT[6]) {
	# Generate results
	results <- data.frame(inputFCS@exprs) %>% 
			mutate(bits = lapply(Sort.Classifier, function(x) as.numeric(intToBits(x))),
					IdxRow = sapply(bits, function(x) sum(x[27], x[28]*2, x[29]*4 + x[30]*8 + x[31]*16 + x[32]*32)),
					IdxCol = sapply(bits, function(x) sum(x[21], x[22]*2, x[23]*4 + x[24]*8 + x[25]*16 + x[26]*32))) %>%
			mutate(IdxRow = rawToChar(as.raw(64 + as.integer(.[, "IdxRow"])), multiple = TRUE)) %>%
			mutate(IdxCol = formatC((IdxCol), width = 2, flag = 0)) %>%
			mutate(IdxRow = ifelse(IdxRow == "@", NA_real_,  IdxRow)) %>%
			mutate(IdxCol = ifelse(IdxCol == "00", NA_real_,  IdxCol)) %>%
			select("IdxRow", "IdxCol")
	
	# Generate perCell
	perCell <- cbind(data.frame(inputFCS@exprs), results) %>%
			filter(IdxRow != "NA") %>%
			mutate(Well = paste0(IdxRow, IdxCol))
}

##### Outputs
# Generate perWell
perWell <- perCell %>%
		group_by(Well, IdxRow, IdxCol) %>%
		mutate(nbrCell = length(Well)) %>%
		summarise_if(is.numeric, median)

# Plate template
plateX <- ifelse(max(as.numeric(perWell$IdxCol)) %in% c(1:3), 3,
		ifelse(max(as.numeric(perWell$IdxCol)) %in% c(4:6), 6,
				ifelse(max(as.numeric(perWell$IdxCol)) %in% c(7:8), 8,
						ifelse(max(as.numeric(perWell$IdxCol)) %in% c(9:12), 12, 24))))
plateY <- ifelse(plateX == 3, 2, ifelse(plateX == 6, 4, ifelse(plateX == 8, 6, ifelse(plateX == 12, 8, 16))))
plateTem <- data.frame(IdxRow = unlist(lapply(LETTERS[1:plateY], function(x) rep(x, plateX))),
		IdxCol = rep(1:plateX, plateY))

# Write csv
write.csv(results, file = "FJ_CSV_OUPUT_FILE", row.names = FALSE, quote = FALSE)
write.csv(perCell, file = paste0("FJ_CSV_OUPUT_FILE", "_perCell.csv"), row.names = FALSE, quote = FALSE)
write.csv(perWell, file = paste0("FJ_CSV_OUPUT_FILE", "_perWell.csv"), row.names = FALSE, quote = FALSE)

# Plot plate
png(file = paste0("FJ_CSV_OUPUT_FILE", ".png"), bg = "transparent", 
		width = 800, height = 520, units = "px", pointsize = 24)
perWell %>%
		ggplot(aes(y = IdxRow, x = as.integer(IdxCol))) + 
		geom_point(shape = 1, size = 12, data = plateTem) +
		geom_point(aes(colour = nbrCell), size = 10, alpha = .75) +
		geom_text(aes(label = nbrCell, alpha = .25)) +
		scale_y_discrete(limits = rev(LETTERS[1:plateY])) +
		scale_x_continuous(breaks = seq(1, plateX)) +
		theme_bw() +
		theme(legend.position = "none",
				panel.grid.minor = element_blank()) +
		labs(x = NULL, y = NULL,
				title = "Number of Cells per Well",
				subtitle = paste0("FJ_CSV_OUPUT_FILE"))
dev.off()
